(* Codice completo per semplice interprete iterativo di espressioni intere *)

(* Definizione concreta di stack, polimorfo: 'a stack *)

type 'a stack = Empty | Push of 'a stack * 'a
     let emptystack = Empty
     let isempty p = (p = emptystack)
     let push p a = Push(p,a) 
     let pop p = match p with
   	| Push(p1, _) -> p1
		| Empty -> failwith "pop on empty stack"
     let top p = match p with
   	| Push(_, a) -> a   
		| Empty -> failwith "top on empty stack"
   
(* Definizione concreta di ambiente, polimorfo: 't env *)

type 't env = (string * 't) list

let rec lookup ide (amb: 't env) = match amb with
   | (ide1, val1) :: amb1 -> if ide = ide1 then val1 else (lookup ide amb1)
   | [] -> failwith  (" "^ide^" not bound in environment");;

(* Espressioni intere *)

type expr = 
  | CstI of int             (* costanti  *)
  | Var of string           (* variabili *)
  | Prim of string * expr * expr  (* operatore binario: consideriamo *,+,- *)

(* Espressioni colorate, come nell'esempio di animazione *)

type coloredExp = 
   | Red of expr      (* Espressione ancora da valutare *)
   | Blue of expr     (* Continuazione, i.e. espressione in attesa dei valori delle sottoespressioni, per applicarci l'operatore *)

(* Stack di servizio: continuazioni e valori temporanei di tipo intero *)

let contSt = ref Empty;;

let tempSt =  ref Empty;;

(* Interprete iterativo *)

let evalIt ((e:expr), (rho: int env)) = 
   contSt := push !contSt (Red(e));
   while not(isempty(!contSt)) do  
      match top(!contSt) with
         | Red(x) -> 
            (contSt :=  pop !contSt; 
            match x with  
       	       | CstI a -> tempSt := push !tempSt a;
       	       | Var str -> tempSt := push !tempSt (lookup str rho);
               | Prim (op, e1, e2) -> 
	          (contSt := push !contSt (Blue(x)); 
	          contSt := push !contSt (Red(e1));
	          contSt := push !contSt (Red(e2)))
            )
         | Blue(x) -> 
            (contSt :=  pop !contSt; 
            match x with 
               | Prim (op, _, _) -> 
                  let e1 = top !tempSt in
                     tempSt := pop !tempSt;
                     let e2 = top !tempSt in
                        (tempSt := pop !tempSt;
                        match op with 
                        |"+" -> tempSt := push !tempSt (e1 + e2)
                        |"-" -> tempSt := push !tempSt (e1 - e2)
                        |"*" -> tempSt := push !tempSt (e1 * e2)
                        | str -> failwith (" unknown operator "^str)
                        )
               | _ -> failwith (" erroneous blue expression")
           )
    done;
    let res = top !tempSt in tempSt := pop !tempSt; res ;;

(* Esempio di valutazione: l'espressione dell'animazione *)
(*   +( *(X,2),-(Y,3)) nell’ambiente {X -> 5, Y -> 7}   *)

(* L'espressione *)
let myExpr = Prim ("+", Prim("*", Var("X"), CstI(2)), Prim("-", Var("Y"), CstI(3)));;

(* L'ambiente *)
let myEnv =  [("X",5);("Y",7)];;

(* La valutazione: ci aspettiamo 14 *)

evalIt(myExpr, myEnv);;

(* Test per vedere errori *)
(* Variable non bound *)
evalIt(Prim("*", Var("Z"), CstI(2)),[("X",5);("Y",7)])
(* Operatore non conosciuto *)
evalIt(Prim("/", Var("X"), CstI(2)),[("X",5);("Y",7)])
