type texpr =                            (* target expressions *)
  | TCstI of int
  | TVar of int                         (* index into runtime environment *)
  | TLet of texpr * texpr               (* erhs and ebody                 *)
  | TPrim of string * texpr * texpr

(* Legame variabile indice a compilazione *)

let rec getindex env x = 
    match env with 
    | [] -> failwith("Variable not found")
    | y::yr -> if x=y then 0 else 1 + getindex yr x

* Compilare Expr in TExpr *)

let rec tcomp e (cenv : string list) : texpr =
    match e with
      | CstI i -> TCstI i
      | Var x  -> TVar (getindex cenv x)
      | Let(x, erhs, ebody) -> 
            let cenv1 = x :: cenv in
            TLet(tcomp erhs cenv, tcomp ebody cenv1)
      | Prim(ope, e1, e2) -> TPrim(ope, tcomp e1 cenv, tcomp e2 cenv)

(* Interprete Text *)
(* Ambiente a run time e' una lista di interi *)

open list;;

let rec teval (e : texpr) (renv : int list) : int =
    match e with
      | TCstI i -> i
      | TVar x  -> nth renv x
      | TLet(erhs, ebody) -> 
            let xval = teval erhs renv in
               let renv1 = xval :: renv  in
                  teval ebody renv1
      | TPrim("+", e1, e2) -> teval e1 renv + teval e2 renv
      | TPrim("*", e1, e2) -> teval e1 renv * teval e2 renv
      | TPrim("-", e1, e2) -> teval e1 renv - teval e2 renv
      | TPrim _            ->failwith("unknown primitive")

let e2 = Let("z", CstI 17, 
             Prim("+", Let("z", CstI 22, Prim("*", CstI 100, Var "z")),
                       Var "z"));;

(* Correctness: eval e [] equals teval (tcomp e []) [] *)
